program main

  implicit none
  include 'mpif.h'

  integer(4) :: myrank,nprocs,ierr,status(MPI_STATUS_SIZE)
  real(8), allocatable :: a(:),b(:),c(:),b_exact(:)
  integer(4) :: irow,ifile,tmp,i,icol
  character(len=50) :: filename
  integer(4) :: nrows,ncols,nnz
  integer(4), allocatable :: row(:),col(:),ptr(:),col_of_row(:)
  integer(4), allocatable :: to_send(:)
  integer(4) :: nto_send, nto_recv
  real(8), allocatable :: buffer_to_send(:)
  integer(4) :: pto_sendrecv
  real(8), parameter :: weird_no=-9191.9191d0
  real(8) :: maxdiff
  logical :: is_print_c=.false.

  call mpi_init(ierr)
  call mpi_comm_rank(mpi_comm_world,myrank,ierr)
  call mpi_comm_size(mpi_comm_world,nprocs,ierr)

  if(nprocs .ne. 2) stop " -- Run with just 2 processors."

  nrows=12 ! 12 rows within each processor.
  ! ID of the 12 rows.
  allocate(row(nrows)); row(:)=-1  
  do irow=1,nrows
     row(irow)=myrank*nrows + irow
  enddo

  ! Column ids.
  write(filename,"(a,i1,a)") "col_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading columns:'
  ifile=100
  open(unit=ifile,file=trim(filename),action='read')
  read(ifile,*) ncols
  allocate(col(ncols)); col(:)=-1
  do i=1,ncols
     read(ifile,*) col(i)
  enddo
  close(ifile)
  if(any(col==-1)) stop " -- Incorrect col. Exiting."

  ! Pointer array for Compressed Row Storage format.
  allocate(ptr(nrows+1)); ptr(:)=-1
  ifile=100
  write(filename,"(a,i1,a)") "ptr_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading pointer array for Compressed Row Storage format:'
  open(unit=ifile,file=trim(filename),action='read')
  do irow=1,nrows+1
     read(ifile,*) ptr(irow)
  enddo
  close(ifile)
  if(any(ptr==-1)) stop " -- Incorrect ptr. Exiting."

  ! Number of non-zeros.
  nnz=ptr(nrows+1)-1

  ! Non-zero columns in each row.
  allocate(col_of_row(nnz)); col_of_row(:)=-1
  ifile=100
  write(filename,"(a,i1,a)") "col_of_row_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading non-zero columns of each row:'
  open(unit=ifile,file=trim(filename),action='read')
  do i=1,nnz
     read(ifile,*) col_of_row(i)
  enddo
  close(ifile)
  if(any(col_of_row==-1)) stop " -- Incorrect col_of_row. Exiting."

  ! Read non-zero entries of matrix A.
  allocate(a(nnz)); a(:)=weird_no
  ifile=100
  write(filename,"(a,i1,a)") "A_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading non-zero entries of matrix A:'
  open(unit=ifile,file=trim(filename),action='read')
  do i=1,nnz
     read(ifile,*) a(i)
  enddo
  close(ifile)
  if(any(a==weird_no)) stop " -- Incorrect A. Exiting."

  ! Read the portion of c to be stored in the current processor.
  allocate(c(ncols)); c(:)=weird_no
  ifile=100
  write(filename,"(a,i1,a)") "c_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading the portion of c to be stored in the current processor:'
  open(unit=ifile,file=trim(filename),action='read')
  do i=1,nrows
     read(ifile,*) c(i)
  enddo
  close(ifile)
  if(any(c(1:nrows)==weird_no)) stop " -- Incorrect c. Exiting."

  ! Metadata for interprocessor communication.
  ifile=100
  write(filename,"(a,i1,a)") "metadata_to_send_", myrank,".dat"
  if(myrank == 0) write(*,*) 'Reading metadata for interprocessor communication:'
  open(unit=ifile,file=trim(filename),action='read')
  read(ifile,*) pto_sendrecv
  read(ifile,*) nto_send
  allocate(to_send(nto_send)); to_send(:)=-1
  do i=1,nto_send
     read(ifile,*) to_send(i)
  enddo
  close(ifile)
  if(any(to_send==-1)) stop " -- Incorrect to_send. Exiting."

  ! Number of c's to receive.
  nto_recv=ncols-nrows
  

  if(myrank==0) write(*,*) 'Finished reading all the data: '

  ! Exchange data.
  if(myrank==0) write(*,*) "Communicating ... "
  allocate(buffer_to_send(nto_send))
  do i=1,nto_send
     buffer_to_send(i)=c(to_send(i))
  enddo
  if(myrank==0) then
     call mpi_send (buffer_to_send, nto_send, MPI_DOUBLE_PRECISION, pto_sendrecv, 0, mpi_comm_world, ierr)
     call mpi_recv (c(nrows+1), nto_recv, MPI_DOUBLE_PRECISION, pto_sendrecv, 0, mpi_comm_world, status, ierr)
  else
     call mpi_recv (c(nrows+1), nto_recv, MPI_DOUBLE_PRECISION, pto_sendrecv, 0, mpi_comm_world, status, ierr)     
     call mpi_send (buffer_to_send, nto_send, MPI_DOUBLE_PRECISION, pto_sendrecv, 0, mpi_comm_world, ierr)
  end if
  deallocate(buffer_to_send)

  if(myrank==0) write(*,*) "Finished communicating "  

  ! Perform matrix-vector multiplication.
  if(myrank==0) write(*,*) "Performing matrix-vector multiplication "  
  allocate(b(nrows)); b(:)=0d0
  do irow=1,nrows
     do i=ptr(irow),ptr(irow+1)-1
        b(irow)=b(irow)+a(i)*c(col_of_row(i))
     enddo
  enddo

  ! Read b_exact.
  allocate(b_exact(nrows)); b_exact(:)=weird_no
  ifile=100
  write(filename,"(a,i1,a)") "b_exact_", myrank,".dat"
  !!  if(myrank == 0) write(*,*) 'Reading b_exact:'
  open(unit=ifile,file=trim(filename),action='read')
  do irow=1,nrows
     read(ifile,*) b_exact(irow)
  enddo
  close(ifile)
  if(any(b_exact == weird_no)) stop " -- Incorrect b_exact. Exiting."

  ! Compute the difference between b and b_exact.
  call mpi_barrier(mpi_comm_world,ierr)
  maxdiff=maxval(abs(b-b_exact))
  if(myrank==0) then
     write(*,*) '------------'
     write(*,'(a)',advance='no') " Maximum difference between the computed b and b_exact (computed from MATLAB): "
     write(*,*)
     write(*,*) '------------'
  endif    
  do i=1,nprocs
     call mpi_barrier(mpi_comm_world,ierr)
     if(i .ne. myrank+1) cycle
     write(*,'(a,i1,2a,e25.15e3)',advance='no') 'Rank ',myrank,' : ', 'maximum difference', maxdiff
     write(*,*)
  enddo  

  call mpi_finalize(ierr)

end program main
